# 20 "src/parser/lexer.mll"
 
  open Format
  open Lexing
  open Term
  open Ptree
  open Parser

  (* lexical errors *)

  exception IllegalCharacter of char
  exception UnterminatedComment
  exception UnterminatedString
  exception AmbiguousPath of string * string

  let () = Exn_printer.register (fun fmt e -> match e with
    | IllegalCharacter c -> fprintf fmt "illegal character %c" c
    | UnterminatedComment -> fprintf fmt "unterminated comment"
    | UnterminatedString -> fprintf fmt "unterminated string"
    | AmbiguousPath (f1, f2) ->
        fprintf fmt "ambiguous path:@ both `%s'@ and `%s'@ match" f1 f2
    | _ -> raise e)

  let keywords = Hashtbl.create 97
  let () =
    List.iter
      (fun (x,y) -> Hashtbl.add keywords x y)
      [
	"and", AND;
	"as", AS;
	"axiom", AXIOM;
	"clone", CLONE;
	"else", ELSE;
	"end", END;
	"epsilon", EPSILON;
	"exists", EXISTS;
	"export", EXPORT;
        "false", FALSE;
	"forall", FORALL;
	"goal", GOAL;
	"if", IF;
	"import", IMPORT;
	"in", IN;
	"inductive", INDUCTIVE;
	"lemma", LEMMA;
	"let", LET;
	"logic", LOGIC;
	"match", MATCH;
	"meta", META;
	"namespace", NAMESPACE;
	"not", NOT;
	"or", OR;
        "prop", PROP;
	"then", THEN;
	"theory", THEORY;
	"true", TRUE;
	"type", TYPE;
	"use", USE;
	"with", WITH;
	(* programs *)
	"abstract", ABSTRACT;
	"absurd", ABSURD;
	"any", ANY;
	"assert", ASSERT;
	"assume", ASSUME;
	"begin", BEGIN;
        "check", CHECK;
	"do", DO;
	"done", DONE;
 	"downto", DOWNTO;
	"exception", EXCEPTION;
	"for", FOR;
	"fun", FUN;
	(* "ghost", GHOST; *)
	"invariant", INVARIANT;
	"label", LABEL;
	"model", MODEL;
	"module", MODULE;
	"mutable", MUTABLE;
	"parameter", PARAMETER;
	"raise", RAISE;
	"raises", RAISES;
	"reads", READS;
	"rec", REC;
	"to", TO;
	"try", TRY;
	"variant", VARIANT;
	"while", WHILE;
        "writes", WRITES;
      ]

  let newline lexbuf =
    let pos = lexbuf.lex_curr_p in
    lexbuf.lex_curr_p <-
      { pos with pos_lnum = pos.pos_lnum + 1; pos_bol = pos.pos_cnum }

  let string_start_loc = ref Loc.dummy_position
  let string_buf = Buffer.create 1024

  let comment_start_loc = ref Loc.dummy_position

  let char_for_backslash = function
    | 'n' -> '\n'
    | 't' -> '\t'
    | c -> c

  let update_loc lexbuf file line chars =
    let pos = lexbuf.lex_curr_p in
    let new_file = match file with None -> pos.pos_fname | Some s -> s in
    lexbuf.lex_curr_p <-
      { pos with
	  pos_fname = new_file;
	  pos_lnum = int_of_string line;
	  pos_bol = pos.pos_cnum - int_of_string chars;
      }

  let remove_leading_plus s =
    let n = String.length s in
    if n > 0 && s.[0] = '+' then String.sub s 1 (n-1) else s

  let loc lb = (lexeme_start_p lb, lexeme_end_p lb)


# 125 "src/parser/lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\215\255\216\255\217\255\094\000\129\000\164\000\199\000\
    \234\000\223\255\224\255\013\001\100\000\048\001\083\001\118\001\
    \153\001\236\255\188\001\238\255\239\255\240\255\242\255\243\255\
    \244\255\086\000\238\001\209\001\014\002\087\002\171\002\255\002\
    \129\000\254\255\113\003\169\003\134\000\179\003\229\003\239\003\
    \032\004\042\004\255\255\091\004\101\004\173\000\150\004\160\004\
    \255\255\209\004\219\004\208\000\083\003\234\004\086\000\036\002\
    \246\004\013\005\028\005\066\005\099\005\131\005\169\005\196\005\
    \206\005\224\005\145\003\100\000\001\006\033\006\048\006\058\006\
    \080\006\090\006\100\006\112\006\122\006\132\006\154\006\164\006\
    \174\006\102\000\246\255\199\006\234\006\013\007\048\007\083\007\
    \118\007\229\255\230\255\173\000\250\255\251\255\252\255\104\000\
    \104\000\109\000\255\255\254\255\079\001\251\255\252\255\253\255\
    \008\000\255\255\254\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\255\255\037\000\036\000\035\000\034\000\
    \033\000\255\255\255\255\029\000\024\000\028\000\037\000\034\000\
    \035\000\255\255\018\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\014\000\027\000\006\000\006\000\005\000\004\000\003\000\
    \002\000\255\255\037\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\007\000\255\255\255\255\255\255\
    \255\255\006\000\255\255\006\000\255\255\255\255\006\000\008\000\
    \255\255\255\255\006\000\006\000\007\000\255\255\255\255\007\000\
    \255\255\007\000\255\255\007\000\255\255\007\000\255\255\007\000\
    \255\255\010\000\255\255\020\000\030\000\034\000\021\000\022\000\
    \023\000\255\255\255\255\255\255\255\255\255\255\255\255\005\000\
    \005\000\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \004\000\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\036\000\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\092\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\000\000\000\000\101\000\000\000\000\000\000\000\
    \106\000\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\032\000\033\000\000\000\000\000\032\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \032\000\008\000\003\000\034\000\004\000\005\000\014\000\024\000\
    \025\000\021\000\005\000\006\000\022\000\016\000\026\000\005\000\
    \028\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\018\000\017\000\015\000\011\000\007\000\008\000\
    \004\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\010\000\012\000\009\000\004\000\031\000\
    \023\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\020\000\013\000\019\000\007\000\004\000\
    \081\000\004\000\004\000\005\000\004\000\089\000\067\000\067\000\
    \005\000\006\000\032\000\006\000\004\000\005\000\032\000\082\000\
    \255\255\099\000\097\000\255\255\067\000\067\000\098\000\000\000\
    \004\000\000\000\007\000\007\000\007\000\004\000\004\000\000\000\
    \000\000\032\000\005\000\090\000\005\000\005\000\005\000\005\000\
    \038\000\000\000\000\000\005\000\006\000\000\000\006\000\005\000\
    \005\000\000\000\000\000\000\000\000\000\000\000\045\000\094\000\
    \000\000\000\000\045\000\005\000\004\000\007\000\007\000\007\000\
    \005\000\005\000\000\000\067\000\000\000\006\000\000\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\045\000\006\000\006\000\
    \042\000\006\000\006\000\006\000\000\000\096\000\000\000\095\000\
    \000\000\051\000\004\000\000\000\007\000\051\000\006\000\005\000\
    \007\000\007\000\007\000\006\000\006\000\000\000\000\000\000\000\
    \007\000\000\000\007\000\007\000\007\000\007\000\000\000\000\000\
    \051\000\007\000\007\000\048\000\007\000\007\000\007\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\005\000\000\000\007\000\
    \002\000\007\000\006\000\007\000\007\000\007\000\007\000\007\000\
    \255\255\000\000\000\000\008\000\000\000\008\000\008\000\005\000\
    \008\000\000\000\000\000\000\000\005\000\006\000\000\000\006\000\
    \008\000\005\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\007\000\000\000\008\000\007\000\007\000\007\000\
    \007\000\008\000\008\000\000\000\000\000\000\000\007\000\000\000\
    \007\000\007\000\007\000\007\000\000\000\000\000\000\000\007\000\
    \007\000\000\000\007\000\007\000\007\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\007\000\000\000\007\000\000\000\007\000\
    \008\000\007\000\007\000\007\000\007\000\007\000\000\000\000\000\
    \000\000\004\000\000\000\004\000\004\000\005\000\004\000\000\000\
    \000\000\103\000\005\000\006\000\000\000\006\000\004\000\005\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\008\000\000\000\
    \007\000\000\000\004\000\007\000\007\000\007\000\007\000\004\000\
    \004\000\105\000\000\000\000\000\004\000\000\000\004\000\004\000\
    \005\000\087\000\000\000\000\000\000\000\005\000\006\000\000\000\
    \006\000\004\000\005\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\007\000\000\000\007\000\000\000\004\000\004\000\007\000\
    \007\000\007\000\004\000\004\000\000\000\000\000\000\000\007\000\
    \000\000\007\000\007\000\007\000\007\000\000\000\000\000\000\000\
    \007\000\007\000\000\000\085\000\007\000\007\000\000\000\000\000\
    \000\000\000\000\000\000\104\000\088\000\093\000\007\000\000\000\
    \007\000\004\000\007\000\007\000\084\000\007\000\007\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\
    \000\000\007\000\000\000\006\000\007\000\007\000\007\000\083\000\
    \006\000\006\000\000\000\000\000\000\000\004\000\000\000\004\000\
    \004\000\005\000\004\000\000\000\000\000\000\000\005\000\006\000\
    \000\000\006\000\004\000\005\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\007\000\000\000\007\000\000\000\004\000\006\000\
    \007\000\007\000\007\000\004\000\004\000\000\000\000\000\052\000\
    \000\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\000\000\000\000\000\000\000\000\004\000\
    \000\000\004\000\004\000\005\000\004\000\006\000\053\000\007\000\
    \005\000\006\000\004\000\006\000\004\000\005\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \004\000\000\000\007\000\007\000\007\000\004\000\004\000\000\000\
    \057\000\000\000\000\000\000\000\000\000\000\000\053\000\000\000\
    \004\000\000\000\007\000\000\000\052\000\000\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \000\000\000\000\000\000\000\000\004\000\000\000\000\000\102\000\
    \054\000\000\000\000\000\053\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\000\000\055\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\056\000\000\000\
    \000\000\000\000\004\000\000\000\007\000\057\000\000\000\000\000\
    \054\000\000\000\000\000\053\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\055\000\029\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\056\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\000\000\000\000\000\000\000\000\029\000\000\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\030\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\000\000\000\000\
    \000\000\000\000\030\000\000\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\000\000\000\000\000\000\000\000\030\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\037\000\000\000\000\000\000\000\037\000\000\000\
    \000\000\000\000\000\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\000\000\000\000\000\000\
    \000\000\037\000\004\000\036\000\004\000\004\000\005\000\004\000\
    \070\000\000\000\000\000\005\000\006\000\000\000\006\000\004\000\
    \005\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\004\000\000\000\007\000\007\000\007\000\
    \004\000\004\000\046\000\000\000\000\000\000\000\046\000\000\000\
    \070\000\000\000\000\000\000\000\037\000\000\000\000\000\000\000\
    \037\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\046\000\000\000\000\000\000\000\000\000\000\000\004\000\
    \000\000\000\000\000\000\037\000\000\000\036\000\000\000\000\000\
    \000\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\004\000\038\000\007\000\
    \066\000\000\000\038\000\000\000\000\000\000\000\000\000\000\000\
    \040\000\000\000\000\000\000\000\040\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\038\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\040\000\
    \000\000\000\000\000\000\000\000\000\000\039\000\039\000\039\000\
    \039\000\039\000\039\000\039\000\039\000\039\000\039\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\040\000\000\000\000\000\000\000\040\000\000\000\000\000\
    \000\000\000\000\000\000\043\000\000\000\000\000\000\000\043\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \040\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\043\000\000\000\000\000\042\000\000\000\000\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\043\000\000\000\000\000\000\000\
    \043\000\000\000\000\000\000\000\000\000\000\000\045\000\000\000\
    \000\000\000\000\045\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\043\000\000\000\000\000\042\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\045\000\000\000\000\000\
    \042\000\000\000\000\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\046\000\
    \000\000\000\000\000\000\046\000\000\000\000\000\000\000\000\000\
    \000\000\049\000\000\000\000\000\000\000\049\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\046\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \049\000\000\000\000\000\048\000\000\000\000\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\049\000\000\000\000\000\000\000\049\000\000\000\
    \000\000\000\000\000\000\000\000\051\000\000\000\000\000\000\000\
    \051\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\049\000\000\000\000\000\048\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\051\000\000\000\000\000\048\000\000\000\
    \000\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\069\000\000\000\069\000\
    \000\000\000\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\058\000\000\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\000\000\
    \000\000\000\000\000\000\000\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\000\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\057\000\000\000\000\000\000\000\
    \061\000\000\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\000\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\000\000\000\000\000\000\000\000\000\000\064\000\000\000\
    \064\000\000\000\060\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\000\000\000\000\000\000\
    \000\000\062\000\000\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\060\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\060\000\000\000\000\000\000\000\000\000\
    \000\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\000\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \000\000\000\000\000\000\060\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \062\000\000\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\065\000\065\000\065\000\065\000\065\000\065\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \060\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\065\000\065\000\065\000\065\000\065\000\065\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \060\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\076\000\000\000\076\000\000\000\000\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\074\000\000\000\074\000\000\000\072\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\000\000\072\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\080\000\000\000\080\000\
    \000\000\078\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \007\000\078\000\007\000\007\000\007\000\007\000\000\000\000\000\
    \000\000\007\000\007\000\000\000\007\000\007\000\007\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\007\000\000\000\007\000\007\000\007\000\007\000\007\000\
    \000\000\000\000\000\000\007\000\000\000\007\000\007\000\007\000\
    \007\000\000\000\000\000\000\000\007\000\007\000\000\000\007\000\
    \007\000\007\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\000\000\000\000\000\000\007\000\000\000\
    \007\000\007\000\007\000\007\000\000\000\000\000\000\000\007\000\
    \007\000\000\000\007\000\007\000\007\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\007\000\000\000\007\000\000\000\007\000\
    \007\000\007\000\007\000\086\000\007\000\007\000\000\000\000\000\
    \000\000\007\000\000\000\007\000\007\000\007\000\007\000\000\000\
    \000\000\000\000\007\000\007\000\000\000\007\000\007\000\007\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\007\000\000\000\
    \007\000\000\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\000\000\000\000\000\000\004\000\000\000\004\000\004\000\
    \005\000\004\000\000\000\000\000\000\000\005\000\006\000\000\000\
    \006\000\004\000\005\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\007\000\000\000\007\000\000\000\004\000\007\000\007\000\
    \007\000\007\000\004\000\004\000\000\000\000\000\000\000\004\000\
    \000\000\004\000\004\000\005\000\004\000\000\000\000\000\000\000\
    \005\000\006\000\000\000\006\000\004\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\007\000\000\000\007\000\000\000\
    \004\000\004\000\007\000\007\000\007\000\004\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\
    \000\000\007\000\000\000\000\000\004\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\004\000\000\000\007\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\
    \025\000\004\000\004\000\004\000\004\000\012\000\054\000\054\000\
    \004\000\004\000\032\000\004\000\004\000\004\000\032\000\081\000\
    \036\000\095\000\096\000\036\000\067\000\067\000\097\000\255\255\
    \004\000\255\255\004\000\004\000\004\000\004\000\004\000\255\255\
    \255\255\032\000\005\000\012\000\005\000\005\000\005\000\005\000\
    \036\000\255\255\255\255\005\000\005\000\255\255\005\000\005\000\
    \005\000\255\255\255\255\255\255\255\255\255\255\045\000\091\000\
    \255\255\255\255\045\000\005\000\004\000\005\000\005\000\005\000\
    \005\000\005\000\255\255\067\000\255\255\006\000\255\255\006\000\
    \006\000\006\000\006\000\255\255\255\255\045\000\006\000\006\000\
    \045\000\006\000\006\000\006\000\255\255\091\000\255\255\091\000\
    \255\255\051\000\004\000\255\255\004\000\051\000\006\000\005\000\
    \006\000\006\000\006\000\006\000\006\000\255\255\255\255\255\255\
    \007\000\255\255\007\000\007\000\007\000\007\000\255\255\255\255\
    \051\000\007\000\007\000\051\000\007\000\007\000\007\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\005\000\255\255\005\000\
    \000\000\007\000\006\000\007\000\007\000\007\000\007\000\007\000\
    \104\000\255\255\255\255\008\000\255\255\008\000\008\000\008\000\
    \008\000\255\255\255\255\255\255\008\000\008\000\255\255\008\000\
    \008\000\008\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \006\000\255\255\006\000\255\255\008\000\007\000\008\000\008\000\
    \008\000\008\000\008\000\255\255\255\255\255\255\011\000\255\255\
    \011\000\011\000\011\000\011\000\255\255\255\255\255\255\011\000\
    \011\000\255\255\011\000\011\000\011\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\007\000\255\255\007\000\255\255\011\000\
    \008\000\011\000\011\000\011\000\011\000\011\000\255\255\255\255\
    \255\255\013\000\255\255\013\000\013\000\013\000\013\000\255\255\
    \255\255\100\000\013\000\013\000\255\255\013\000\013\000\013\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \008\000\255\255\013\000\011\000\013\000\013\000\013\000\013\000\
    \013\000\100\000\255\255\255\255\014\000\255\255\014\000\014\000\
    \014\000\014\000\255\255\255\255\255\255\014\000\014\000\255\255\
    \014\000\014\000\014\000\255\255\255\255\255\255\036\000\255\255\
    \255\255\011\000\255\255\011\000\255\255\014\000\013\000\014\000\
    \014\000\014\000\014\000\014\000\255\255\255\255\255\255\015\000\
    \255\255\015\000\015\000\015\000\015\000\255\255\255\255\255\255\
    \015\000\015\000\255\255\015\000\015\000\015\000\255\255\255\255\
    \255\255\255\255\255\255\100\000\013\000\091\000\013\000\255\255\
    \015\000\014\000\015\000\015\000\015\000\015\000\015\000\255\255\
    \255\255\255\255\016\000\255\255\016\000\016\000\016\000\016\000\
    \255\255\255\255\255\255\016\000\016\000\255\255\016\000\016\000\
    \016\000\255\255\255\255\255\255\255\255\255\255\255\255\014\000\
    \255\255\014\000\255\255\016\000\015\000\016\000\016\000\016\000\
    \016\000\016\000\255\255\255\255\255\255\018\000\255\255\018\000\
    \018\000\018\000\018\000\255\255\255\255\255\255\018\000\018\000\
    \255\255\018\000\018\000\018\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\015\000\255\255\015\000\255\255\018\000\016\000\
    \018\000\018\000\018\000\018\000\018\000\255\255\255\255\027\000\
    \255\255\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\255\255\255\255\255\255\255\255\026\000\
    \255\255\026\000\026\000\026\000\026\000\016\000\027\000\016\000\
    \026\000\026\000\018\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\255\255\026\000\026\000\026\000\026\000\026\000\255\255\
    \027\000\255\255\255\255\255\255\255\255\255\255\027\000\255\255\
    \018\000\255\255\018\000\255\255\028\000\255\255\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \255\255\255\255\255\255\255\255\026\000\255\255\255\255\100\000\
    \028\000\255\255\255\255\028\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\255\255\028\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\028\000\255\255\
    \255\255\255\255\026\000\255\255\026\000\028\000\255\255\255\255\
    \028\000\255\255\255\255\028\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\028\000\029\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\028\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\255\255\255\255\255\255\255\255\029\000\255\255\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\030\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\255\255\255\255\
    \255\255\255\255\030\000\255\255\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\031\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\255\255\255\255\255\255\255\255\031\000\255\255\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\034\000\255\255\255\255\255\255\034\000\255\255\
    \255\255\255\255\255\255\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\255\255\255\255\255\255\
    \255\255\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \052\000\255\255\255\255\034\000\034\000\255\255\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\255\255\034\000\034\000\034\000\
    \034\000\034\000\035\000\255\255\255\255\255\255\035\000\255\255\
    \052\000\255\255\255\255\255\255\037\000\255\255\255\255\255\255\
    \037\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\035\000\255\255\255\255\255\255\255\255\255\255\034\000\
    \255\255\255\255\255\255\037\000\255\255\037\000\255\255\255\255\
    \255\255\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\034\000\038\000\034\000\
    \066\000\255\255\038\000\255\255\255\255\255\255\255\255\255\255\
    \039\000\255\255\255\255\255\255\039\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\038\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\039\000\
    \255\255\255\255\255\255\255\255\255\255\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\039\000\
    \039\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\
    \039\000\040\000\255\255\255\255\255\255\040\000\255\255\255\255\
    \255\255\255\255\255\255\041\000\255\255\255\255\255\255\041\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \040\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\041\000\255\255\255\255\041\000\255\255\255\255\
    \040\000\040\000\040\000\040\000\040\000\040\000\040\000\040\000\
    \040\000\040\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\043\000\255\255\255\255\255\255\
    \043\000\255\255\255\255\255\255\255\255\255\255\044\000\255\255\
    \255\255\255\255\044\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\043\000\255\255\255\255\043\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\044\000\255\255\255\255\
    \044\000\255\255\255\255\043\000\043\000\043\000\043\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\044\000\044\000\044\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\046\000\
    \255\255\255\255\255\255\046\000\255\255\255\255\255\255\255\255\
    \255\255\047\000\255\255\255\255\255\255\047\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\046\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \047\000\255\255\255\255\047\000\255\255\255\255\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\049\000\255\255\255\255\255\255\049\000\255\255\
    \255\255\255\255\255\255\255\255\050\000\255\255\255\255\255\255\
    \050\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\049\000\255\255\255\255\049\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\050\000\255\255\255\255\050\000\255\255\
    \255\255\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\053\000\255\255\053\000\
    \255\255\255\255\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\056\000\255\255\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\056\000\
    \056\000\056\000\056\000\056\000\056\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\255\255\
    \255\255\255\255\255\255\255\255\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\255\255\056\000\
    \056\000\056\000\056\000\056\000\056\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\057\000\255\255\255\255\255\255\
    \059\000\255\255\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\255\255\058\000\058\000\058\000\
    \058\000\058\000\058\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\255\255\255\255\255\255\255\255\255\255\060\000\255\255\
    \060\000\255\255\059\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\255\255\255\255\255\255\
    \255\255\059\000\255\255\059\000\059\000\059\000\059\000\059\000\
    \059\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\059\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\061\000\061\000\061\000\061\000\
    \061\000\061\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\061\000\255\255\255\255\255\255\255\255\
    \255\255\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\255\255\061\000\061\000\061\000\061\000\
    \061\000\061\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \255\255\255\255\255\255\061\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \062\000\255\255\062\000\062\000\062\000\062\000\062\000\062\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\065\000\065\000\065\000\065\000\065\000\065\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \065\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\065\000\065\000\065\000\065\000\065\000\065\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \065\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\070\000\255\255\070\000\255\255\255\255\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\072\000\255\255\072\000\255\255\071\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\255\255\071\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\078\000\255\255\078\000\
    \255\255\077\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \083\000\077\000\083\000\083\000\083\000\083\000\255\255\255\255\
    \255\255\083\000\083\000\255\255\083\000\083\000\083\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\083\000\255\255\083\000\083\000\083\000\083\000\083\000\
    \255\255\255\255\255\255\084\000\255\255\084\000\084\000\084\000\
    \084\000\255\255\255\255\255\255\084\000\084\000\255\255\084\000\
    \084\000\084\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\084\000\083\000\084\000\084\000\
    \084\000\084\000\084\000\255\255\255\255\255\255\085\000\255\255\
    \085\000\085\000\085\000\085\000\255\255\255\255\255\255\085\000\
    \085\000\255\255\085\000\085\000\085\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\083\000\255\255\083\000\255\255\085\000\
    \084\000\085\000\085\000\085\000\085\000\085\000\255\255\255\255\
    \255\255\086\000\255\255\086\000\086\000\086\000\086\000\255\255\
    \255\255\255\255\086\000\086\000\255\255\086\000\086\000\086\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\084\000\255\255\
    \084\000\255\255\086\000\085\000\086\000\086\000\086\000\086\000\
    \086\000\255\255\255\255\255\255\087\000\255\255\087\000\087\000\
    \087\000\087\000\255\255\255\255\255\255\087\000\087\000\255\255\
    \087\000\087\000\087\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\085\000\255\255\085\000\255\255\087\000\086\000\087\000\
    \087\000\087\000\087\000\087\000\255\255\255\255\255\255\088\000\
    \255\255\088\000\088\000\088\000\088\000\255\255\255\255\255\255\
    \088\000\088\000\255\255\088\000\088\000\088\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\086\000\255\255\086\000\255\255\
    \088\000\087\000\088\000\088\000\088\000\088\000\088\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\087\000\
    \255\255\087\000\255\255\255\255\088\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\088\000\255\255\088\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\010\000\022\000\044\000\000\000\000\000\000\000\
    \000\000\000\000\099\000\125\000\000\000\152\000\178\000\209\000\
    \235\000\005\001\096\000\031\001\016\001\001\000\065\001\080\001\
    \117\000\091\001\103\001\002\000\113\001\032\000\000\000\000\000\
    \135\001\000\000\173\001\198\001\000\000\236\001\000\000\000\000\
    \000\000\018\002\000\000\000\000\123\001\051\002\083\002\093\002\
    \103\002\115\002\125\002\135\002\147\002\157\002\167\002\179\002\
    \189\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\128\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\170\000\
    \000\000\000\000\000\000\000\000\181\000\000\000\000\000\128\000\
    \000\000\181\000\000\000\181\000\000\000\128\000\000\000\181\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\062\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\001\000\112\000\112\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\017\000\000\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\032\000\029\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\001\000\000\000\000\000\000\000\
    \001\000\029\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\029\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\001\000\037\000\042\000\052\000\000\000\
    \000\000\000\000\052\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\029\000\000\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\052\000\000\000\000\000\
    \000\000\001\000\000\000\000\000\037\000\001\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\000\000\
    \001\000\000\000\065\000\001\000\000\000\000\000\000\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\052\000\000\000\000\000\000\000\052\000\000\000\
    \000\000\000\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\000\000\000\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\052\000\000\000\000\000\000\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\052\000\000\000\000\000\082\000\000\000\
    \000\000\000\000\082\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\082\000\000\000\000\000\
    \082\000\000\000\000\000\000\000\082\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\112\000\000\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\082\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\052\000\000\000\000\000\000\000\052\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\082\000\000\000\000\000\000\000\082\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\082\000\000\000\000\000\000\000\
    \082\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \082\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\082\000\000\000\000\000\000\000\000\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\112\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\000\000\000\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\142\000\029\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\029\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\000\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\155\000\000\000\160\000\160\000\
    \160\000\160\000\160\000\160\000\160\000\160\000\160\000\160\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\160\000\
    \160\000\160\000\160\000\160\000\160\000\000\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\000\000\000\000\167\000\000\000\
    \000\000\000\000\000\000\000\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\000\000\160\000\
    \160\000\160\000\160\000\160\000\160\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\000\000\000\000\000\000\167\000\000\000\
    \000\000\000\000\000\000\000\000\167\000\000\000\000\000\000\000\
    \000\000\000\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\000\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\000\000\000\000\000\000\167\000\000\000\000\000\000\000\
    \000\000\000\000\167\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\167\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\000\000\029\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\000\000\029\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\014\000\014\000\014\000\
    \014\000\014\000\014\000\014\000\014\000\014\000\014\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\000\000\029\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\029\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\255\255\255\255\036\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\000\000\045\000\051\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\027\000\255\255\027\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\028\000\027\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\034\000\255\255\255\255\255\255\
    \034\000\028\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\027\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\034\000\028\000\034\000\035\000\255\255\
    \255\255\255\255\035\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\028\000\255\255\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\035\000\255\255\255\255\
    \255\255\037\000\255\255\255\255\028\000\037\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\255\255\
    \037\000\255\255\037\000\038\000\255\255\255\255\255\255\038\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\038\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\039\000\255\255\255\255\255\255\039\000\255\255\
    \255\255\255\255\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\255\255\255\255\255\255\255\255\
    \255\255\039\000\255\255\255\255\040\000\255\255\255\255\255\255\
    \040\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \036\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\
    \039\000\039\000\039\000\040\000\255\255\255\255\041\000\255\255\
    \255\255\255\255\041\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\040\000\040\000\040\000\040\000\040\000\
    \040\000\040\000\040\000\040\000\040\000\041\000\255\255\255\255\
    \043\000\255\255\255\255\255\255\043\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\044\000\255\255\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\043\000\
    \044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\
    \044\000\044\000\046\000\255\255\255\255\255\255\046\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\047\000\255\255\255\255\255\255\047\000\255\255\255\255\
    \255\255\046\000\255\255\255\255\049\000\255\255\255\255\255\255\
    \049\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \047\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\049\000\255\255\255\255\255\255\255\255\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\050\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\255\255\255\255\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\056\000\052\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \056\000\056\000\056\000\056\000\056\000\056\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\052\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\255\255\
    \056\000\056\000\056\000\056\000\056\000\056\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\059\000\255\255\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\059\000\
    \059\000\059\000\059\000\059\000\059\000\255\255\058\000\058\000\
    \058\000\058\000\058\000\058\000\255\255\255\255\059\000\255\255\
    \255\255\255\255\255\255\255\255\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\255\255\059\000\
    \059\000\059\000\059\000\059\000\059\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\255\255\255\255\255\255\059\000\255\255\
    \255\255\255\255\255\255\255\255\061\000\255\255\255\255\255\255\
    \255\255\255\255\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\255\255\061\000\061\000\061\000\
    \061\000\061\000\061\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\255\255\255\255\255\255\061\000\255\255\255\255\255\255\
    \255\255\255\255\065\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\065\000\065\000\065\000\065\000\065\000\
    \065\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\065\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\255\255\071\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\255\255\071\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\255\255\077\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\077\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\010\255\255\008\255\255\009\255\008\255\007\255\255\009\255\
    \255\008\255\009\255\255\007\255\008\255\009\255\255\018\255\255\
    \009\255\008\255\255\020\255\019\255\255\014\255\013\255\255\012\
    \255\011\255\255\012\255\255\017\255\016\255\015\255\255\014\255\
    \255\013\255\014\255\255\011\255\012\255\255\015\011\016\012\017\
    \255\255\016\255\255\011\015\012\016\015\255\016\255\017\255\255\
    \000\010\001\011\005\013\004\014\002\012\003\017\255\017\255\255\
    \011\015\012\016\255\005\255\000\010\001\011\002\012\003\017\255\
    \001\255\005\006\004\007\003\008\002\009\255\024\255\255\021\255\
    \255\022\255\021\255\020\255\255\022\255\255\022\255\021\255\255\
    \020\255\021\255\022\255\255\023\255\255\004\019\003\020\002\021\
    \001\022\000\023\255\005\006\004\007\003\008\002\009\001\018\000\
    \024\255";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.create 25 (-1) ; (* L=2 [7] <- p ; [6] <- p ;  *)
  lexbuf.Lexing.lex_mem.(7) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(6) <- lexbuf.Lexing.lex_curr_pos ;
  __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 174 "src/parser/lexer.mll"
                                                 file
# 1030 "src/parser/lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(5) lexbuf.Lexing.lex_mem.(4)
and
# 175 "src/parser/lexer.mll"
                      line
# 1035 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 175 "src/parser/lexer.mll"
                                              char
# 1040 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3) in
# 176 "src/parser/lexer.mll"
      ( update_loc lexbuf file line char; token lexbuf )
# 1044 "src/parser/lexer.ml"

  | 1 ->
# 178 "src/parser/lexer.mll"
      ( newline lexbuf; token lexbuf )
# 1049 "src/parser/lexer.ml"

  | 2 ->
# 180 "src/parser/lexer.mll"
      ( token lexbuf )
# 1054 "src/parser/lexer.ml"

  | 3 ->
# 182 "src/parser/lexer.mll"
      ( UNDERSCORE )
# 1059 "src/parser/lexer.ml"

  | 4 ->
let
# 183 "src/parser/lexer.mll"
              id
# 1065 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 184 "src/parser/lexer.mll"
      ( try Hashtbl.find keywords id with Not_found -> LIDENT id )
# 1069 "src/parser/lexer.ml"

  | 5 ->
let
# 185 "src/parser/lexer.mll"
              id
# 1075 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 186 "src/parser/lexer.mll"
      ( UIDENT id )
# 1079 "src/parser/lexer.ml"

  | 6 ->
let
# 187 "src/parser/lexer.mll"
                   s
# 1085 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 188 "src/parser/lexer.mll"
      ( INTEGER s )
# 1089 "src/parser/lexer.ml"

  | 7 ->
let
# 189 "src/parser/lexer.mll"
               i
# 1095 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(5) lexbuf.Lexing.lex_mem.(4)
and
# 189 "src/parser/lexer.mll"
                         f
# 1100 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(2)
and
# 189 "src/parser/lexer.mll"
                                                            e
# 1105 "src/parser/lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_mem.(0) in
# 192 "src/parser/lexer.mll"
      ( FLOAT (RConstDecimal (i, f, Util.option_map remove_leading_plus e)) )
# 1109 "src/parser/lexer.ml"

  | 8 ->
let
# 193 "src/parser/lexer.mll"
                                  i
# 1115 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(4) lexbuf.Lexing.lex_mem.(3)
and
# 193 "src/parser/lexer.mll"
                                                        f
# 1120 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(1)
and
# 196 "src/parser/lexer.mll"
                                    e
# 1125 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 197 "src/parser/lexer.mll"
      ( FLOAT (RConstHexa (i, f, remove_leading_plus e)) )
# 1129 "src/parser/lexer.ml"

  | 9 ->
# 199 "src/parser/lexer.mll"
      ( LEFTPAR_STAR_RIGHTPAR )
# 1134 "src/parser/lexer.ml"

  | 10 ->
# 201 "src/parser/lexer.mll"
      ( comment_start_loc := loc lexbuf; comment lexbuf; token lexbuf )
# 1139 "src/parser/lexer.ml"

  | 11 ->
# 203 "src/parser/lexer.mll"
      ( QUOTE )
# 1144 "src/parser/lexer.ml"

  | 12 ->
# 205 "src/parser/lexer.mll"
      ( BACKQUOTE )
# 1149 "src/parser/lexer.ml"

  | 13 ->
# 207 "src/parser/lexer.mll"
      ( COMMA )
# 1154 "src/parser/lexer.ml"

  | 14 ->
# 209 "src/parser/lexer.mll"
      ( LEFTPAR )
# 1159 "src/parser/lexer.ml"

  | 15 ->
# 211 "src/parser/lexer.mll"
      ( RIGHTPAR )
# 1164 "src/parser/lexer.ml"

  | 16 ->
# 213 "src/parser/lexer.mll"
      ( LEFTBRC )
# 1169 "src/parser/lexer.ml"

  | 17 ->
# 215 "src/parser/lexer.mll"
      ( RIGHTBRC )
# 1174 "src/parser/lexer.ml"

  | 18 ->
# 217 "src/parser/lexer.mll"
      ( COLON )
# 1179 "src/parser/lexer.ml"

  | 19 ->
# 219 "src/parser/lexer.mll"
      ( SEMICOLON )
# 1184 "src/parser/lexer.ml"

  | 20 ->
# 221 "src/parser/lexer.mll"
      ( ARROW )
# 1189 "src/parser/lexer.ml"

  | 21 ->
# 223 "src/parser/lexer.mll"
      ( LRARROW )
# 1194 "src/parser/lexer.ml"

  | 22 ->
# 225 "src/parser/lexer.mll"
      ( AMPAMP )
# 1199 "src/parser/lexer.ml"

  | 23 ->
# 227 "src/parser/lexer.mll"
      ( BARBAR )
# 1204 "src/parser/lexer.ml"

  | 24 ->
# 229 "src/parser/lexer.mll"
      ( LAMBDA )
# 1209 "src/parser/lexer.ml"

  | 25 ->
# 231 "src/parser/lexer.mll"
      ( PRED )
# 1214 "src/parser/lexer.ml"

  | 26 ->
# 233 "src/parser/lexer.mll"
      ( FUNC )
# 1219 "src/parser/lexer.ml"

  | 27 ->
# 235 "src/parser/lexer.mll"
      ( DOT )
# 1224 "src/parser/lexer.ml"

  | 28 ->
# 237 "src/parser/lexer.mll"
      ( BAR )
# 1229 "src/parser/lexer.ml"

  | 29 ->
# 239 "src/parser/lexer.mll"
      ( EQUAL )
# 1234 "src/parser/lexer.ml"

  | 30 ->
# 241 "src/parser/lexer.mll"
      ( LTGT )
# 1239 "src/parser/lexer.ml"

  | 31 ->
# 243 "src/parser/lexer.mll"
      ( LEFTSQ )
# 1244 "src/parser/lexer.ml"

  | 32 ->
# 245 "src/parser/lexer.mll"
      ( RIGHTSQ )
# 1249 "src/parser/lexer.ml"

  | 33 ->
let
# 246 "src/parser/lexer.mll"
                               s
# 1255 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 247 "src/parser/lexer.mll"
      ( OPPREF s )
# 1259 "src/parser/lexer.ml"

  | 34 ->
let
# 248 "src/parser/lexer.mll"
                                             s
# 1265 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 249 "src/parser/lexer.mll"
      ( OP1 s )
# 1269 "src/parser/lexer.ml"

  | 35 ->
let
# 250 "src/parser/lexer.mll"
                                             s
# 1275 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 251 "src/parser/lexer.mll"
      ( OP2 s )
# 1279 "src/parser/lexer.ml"

  | 36 ->
let
# 252 "src/parser/lexer.mll"
                                            s
# 1285 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 253 "src/parser/lexer.mll"
      ( OP3 s )
# 1289 "src/parser/lexer.ml"

  | 37 ->
let
# 254 "src/parser/lexer.mll"
                  s
# 1295 "src/parser/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 255 "src/parser/lexer.mll"
      ( OP4 s )
# 1299 "src/parser/lexer.ml"

  | 38 ->
# 257 "src/parser/lexer.mll"
      ( string_start_loc := loc lexbuf; STRING (string lexbuf) )
# 1304 "src/parser/lexer.ml"

  | 39 ->
# 259 "src/parser/lexer.mll"
      ( EOF )
# 1309 "src/parser/lexer.ml"

  | 40 ->
let
# 260 "src/parser/lexer.mll"
         c
# 1315 "src/parser/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 261 "src/parser/lexer.mll"
      ( raise (IllegalCharacter c) )
# 1319 "src/parser/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and comment lexbuf =
    __ocaml_lex_comment_rec lexbuf 91
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 265 "src/parser/lexer.mll"
      ( comment lexbuf )
# 1330 "src/parser/lexer.ml"

  | 1 ->
# 267 "src/parser/lexer.mll"
      ( () )
# 1335 "src/parser/lexer.ml"

  | 2 ->
# 269 "src/parser/lexer.mll"
      ( comment lexbuf; comment lexbuf )
# 1340 "src/parser/lexer.ml"

  | 3 ->
# 271 "src/parser/lexer.mll"
      ( newline lexbuf; comment lexbuf )
# 1345 "src/parser/lexer.ml"

  | 4 ->
# 273 "src/parser/lexer.mll"
      ( raise (Loc.Located (!comment_start_loc, UnterminatedComment)) )
# 1350 "src/parser/lexer.ml"

  | 5 ->
# 275 "src/parser/lexer.mll"
      ( comment lexbuf )
# 1355 "src/parser/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and string lexbuf =
    __ocaml_lex_string_rec lexbuf 100
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 279 "src/parser/lexer.mll"
      ( let s = Buffer.contents string_buf in
	Buffer.clear string_buf;
	s )
# 1368 "src/parser/lexer.ml"

  | 1 ->
let
# 282 "src/parser/lexer.mll"
               c
# 1374 "src/parser/lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
# 283 "src/parser/lexer.mll"
      ( Buffer.add_char string_buf (char_for_backslash c); string lexbuf )
# 1378 "src/parser/lexer.ml"

  | 2 ->
# 285 "src/parser/lexer.mll"
      ( newline lexbuf; Buffer.add_char string_buf '\n'; string lexbuf )
# 1383 "src/parser/lexer.ml"

  | 3 ->
# 287 "src/parser/lexer.mll"
      ( raise (Loc.Located (!string_start_loc, UnterminatedString)) )
# 1388 "src/parser/lexer.ml"

  | 4 ->
let
# 288 "src/parser/lexer.mll"
         c
# 1394 "src/parser/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 289 "src/parser/lexer.mll"
      ( Buffer.add_char string_buf c; string lexbuf )
# 1398 "src/parser/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_string_rec lexbuf __ocaml_lex_state

;;

# 291 "src/parser/lexer.mll"
 
  let with_location f lb =
    if Debug.test_flag Debug.stack_trace then f lb else
    try f lb with
      | Loc.Located _ as e -> raise e
      | e -> raise (Loc.Located (loc lb, e))

  let parse_logic_file env = with_location (logic_file_eof env token)

  let parse_list0_decl env lenv uc =
    with_location (list0_decl_eof env lenv uc token)

  let parse_lexpr = with_location (lexpr_eof token)

  let parse_program_file = with_location (program_file token)

  let read_channel env file c =
    let lb = Lexing.from_channel c in
    Loc.set_file file lb;
    parse_logic_file env lb

  (* searches file [f] in loadpath [lp] *)
  let locate_file lp f =
    let fl = List.map (fun dir -> Filename.concat dir f) lp in
    match List.filter Sys.file_exists fl with
      | [] -> raise Not_found
      | [file] -> file
      | file1 :: file2 :: _ -> raise (AmbiguousPath (file1, file2))

  let create_env lp =
    let ret_chan sl = 
      let f = List.fold_left Filename.concat "" sl in
      let file = locate_file lp f in
      file, open_in file
    in
    let retrieve env sl =
      let f = List.fold_left Filename.concat "" sl ^ ".why" in
      let file = locate_file lp f in
      let c = open_in file in
      try
	let tl = read_channel env file c in
	close_in c;
	tl
      with
	| e -> close_in c; raise e
    in
    Env.create_env ret_chan retrieve

  let () = Env.register_format "why" ["why"] read_channel

# 1455 "src/parser/lexer.ml"
